/******************************************************************************\
 *  _   ___     ____  __               _                                      *
 * | | | \ \___/ /  \/  | ___ __ _ ___| |_                                    *
 * | |_| |\     /| |\/| |/ __/ _` / __| __|                                   *
 * |  _  | \ - / | |  | | (_| (_| \__ \ |_                                    *
 * |_| |_|  \_/  |_|  |_|\___\__,_|___/\__|                                   *
 *                                                                            *
 * This file is part of the HAMcast project.                                  *
 *                                                                            *
 * HAMcast is free software: you can redistribute it and/or modify            *
 * it under the terms of the GNU Lesser General Public License as published   *
 * by the Free Software Foundation, either version 3 of the License, or       *
 * (at your option) any later version.                                        *
 *                                                                            *
 * HAMcast is distributed in the hope that it will be useful,                 *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of             *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                       *
 * See the GNU Lesser General Public License for more details.                *
 *                                                                            *
 * You should have received a copy of the GNU Lesser General Public License   *
 * along with HAMcast. If not, see <http://www.gnu.org/licenses/>.            *
 *                                                                            *
 * Contact: HAMcast support <hamcast-support@informatik.haw-hamburg.de>       *
\******************************************************************************/

#ifndef HAMCAST_CONFIG_HPP
#define HAMCAST_CONFIG_HPP

#include <string>
#include <boost/cstdint.hpp>

#ifdef HC_DOCUMENTATION
#define HAMCAST_MACOS
#define HAMCAST_LINUX
#define HAMCAST_WINDOWS
#define HAMCAST_64BIT
#endif

/**
 * @def HAMCAST_MACOS
 * @brief This macro is defined, if the host is running Mac OS X.
 * @ingroup Config
 *
 * @def HAMCAST_LINUX
 * @brief This macro is defined, if the host is running Linux.
 * @ingroup Config
 *
 * @def HAMCAST_WINDOWS
 * @brief This macro is defined, if the host is running Windows.
 * @ingroup Config

 * @def HAMCAST_64BIT
 * @brief This macro is defined, if the host is running a 64-bit system.
 * @ingroup Config
 */

#if defined(__APPLE__)
#  define HAMCAST_MACOS
#elif defined(__GNUC__) && defined(__linux__)
#  define HAMCAST_LINUX
#elif defined(WIN32)
#  define HAMCAST_WINDOWS
#else
#  error Platform and/or compiler not supportet
#endif

#if defined(__amd64__) || defined(__LP64__)
#  define HAMCAST_64BIT
#endif

namespace hamcast {

/**
 * @brief The HAMcast magic number.
 * @ingroup Config
 */
static const boost::uint32_t magic_number = 0xDEADC0DE;

/**
 * @brief The major version of the library.
 * @ingroup Config
 */
static const boost::uint32_t major_version = 0;

/**
 * @brief The minor version of the library.
 * @ingroup Config
 */
static const boost::uint32_t minor_version = 4;

/**
 * @brief The minimum major version that is (IPC) compatible to this version.
 * @ingroup Config
 */
static const boost::uint32_t min_compatile_major_version = 0;

/**
 * @brief The minimum minor version that is (IPC) compatible to this version.
 * @ingroup Config
 */
static const boost::uint32_t min_compatile_minor_version = 3;

/**
 * @brief The default maximum buffer size (16MB).
 * @ingroup Config
 */
static const size_t default_max_buffer_size = 16 * 1024 * 1024;

/**
 * @brief The default chunk/block size for IO buffers.
 * @note This constant is used by {@link util::buffered_sink}.
 * @ingroup Config
 */
static const size_t default_block_size = 512;

/**
 * @brief The maximum number of pending asynchronous send messages.
 * @ingroup Config
 */
static const size_t max_pending_sends = 256;

/**
 * @brief The maximum number of buffered asynchronous send messages.
 * @ingroup Config
 * @note Should be factor of default_max_buffer_size.
 */
static const size_t max_buffered_sends = 512;

/**
 * @brief Deduced from default_max_buffer_size and max_buffered_sends.
 * @ingroup Config
 */
static const size_t min_buffer_chunk = default_max_buffer_size / max_buffered_sends;

/**
 * @brief The maximum number of cumulative acknowledged messages.
 * @ingroup Config
 */
static const size_t max_ack_block_size = 64;

/**
 * @brief The time interval of forced acknowledgements in microseconds.
 *
 * Default is 5ms.
 * @ingroup Config
 */
static const size_t force_ack_us_interval = 5000;

/**
 * @brief The default maximum size of for IO buffers (5MB).
 * @ingroup Config
 * @note This constant is used by {@link util::buffered_sink}.
 */
static const size_t default_max_write_buffer_size = (1024 * 1024 * 5);

//TODO: distinguish between POSIX and Windows
/**
 * @brief The path of the meeting point.
 *
 * The meeting point is a directory where the middleware
 * stores all config parameters needed by client applications.
 * @ingroup Config
 */
static const char* meeting_point = "/tmp/hamcast/meeting_point/middleware/";

/**
 * @brief The filename of the middleware lock file.
 * @ingroup Config
 */
static const char* lock_filename = "middleware.lock_file";

/**
 * @brief The filename of the middleware config file.
 * @ingroup Config
 */
static const char* config_filename = "middleware.config_file";

/**
 * @brief Utility function to get the library version as string in the format
 *        <code>{major_version}.{minor_version}</code>.
 * @ingroup Config
 * @return The HAMcast version as string.
 */
std::string version_string();

} // namespace hamcast

#endif // HAMCAST_CONFIG_HPP
