/******************************************************************************\
 *  _   ___     ____  __               _                                      *
 * | | | \ \___/ /  \/  | ___ __ _ ___| |_                                    *
 * | |_| |\     /| |\/| |/ __/ _` / __| __|                                   *
 * |  _  | \ - / | |  | | (_| (_| \__ \ |_                                    *
 * |_| |_|  \_/  |_|  |_|\___\__,_|___/\__|                                   *
 *                                                                            *
 * This file is part of the HAMcast project.                                  *
 *                                                                            *
 * HAMcast is free software: you can redistribute it and/or modify            *
 * it under the terms of the GNU Lesser General Public License as published   *
 * by the Free Software Foundation, either version 3 of the License, or       *
 * (at your option) any later version.                                        *
 *                                                                            *
 * HAMcast is distributed in the hope that it will be useful,                 *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of             *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                       *
 * See the GNU Lesser General Public License for more details.                *
 *                                                                            *
 * You should have received a copy of the GNU Lesser General Public License   *
 * along with HAMcast. If not, see <http://www.gnu.org/licenses/>.            *
 *                                                                            *
 * Contact: HAMcast support <hamcast-support@informatik.haw-hamburg.de>       *
\******************************************************************************/

#include <boost/thread.hpp>

#include "hamcast/uri.hpp"
#include "hamcast/ipc/channel.hpp"
#include "hamcast/multicast_socket.hpp"
#include "hamcast/util/id_generator.hpp"
#include "hamcast/util/serialization.hpp"

namespace hamcast { namespace ipc {

void channel::set_err_str(const char* cstr)
{
    boost::mutex::scoped_lock guard(m_err_str_mtx);
    // do only set the first occured error
    if (m_err_str.empty()) m_err_str = cstr;
}

void channel::run_send_loop(channel* self_ptr)
{
    HC_REQUIRE(self_ptr != 0);
    channel::ptr scoped_ptr(self_ptr);
    HC_MEMORY_BARRIER();
    channel& self = *self_ptr;
    // handshake on startup
    // ensures that both threads incremented reference count of self_ptr
    self.m_barrier.wait();
    try
    {
        self.send_loop();
    }
    catch (std::exception& e)
    {
        self.set_err_str(e.what());
    }
    catch (...)
    {
        self.set_err_str("Unknown Exception in send_loop.");
    }
    // this should kill receive_loop with an ios_base::failure
    self.close();
    // wait for receive_loop
    self.m_barrier.wait();
    // receive_loop calls on_exit() now
}

void channel::close()
{
    m_source->close();
    m_sink->close();
}

void channel::run_receive_loop(channel* self_ptr)
{
    HC_REQUIRE(self_ptr != 0);
    channel::ptr scoped_ptr(self_ptr);
    HC_MEMORY_BARRIER();
    channel& self = *self_ptr;
    // handshake on startup
    // ensures that both threads incremented reference count of self_ptr
    self.m_barrier.wait();
    try
    {
        self.receive_loop();
    }
    catch (std::exception& e)
    {
        self.set_err_str(e.what());
    }
    catch (...)
    {
        self.set_err_str("Unknown Exception in receive_loop.");
    }
    // this should kill send_loop with an ios_base::failure
    self.close();
    // wait for send_loop to exit also
    self.m_barrier.wait();
    // run on_exit
    self.on_exit(self.m_err_str);
}

void channel::on_exit(const std::string&)
{
}

channel::channel(const util::source::ptr& in, const util::sink::ptr& out)
    : m_sink(out), m_source(in), m_barrier(2)
{
}

channel::channel(const std::pair<util::source::ptr, util::sink::ptr>& in_out)
    : m_sink(in_out.second), m_source(in_out.first), m_barrier(2)
{
}

void channel::run()
{
    m_send_loop = boost::thread(boost::bind(&channel::run_send_loop, this));
    m_receive_loop = boost::thread(boost::bind(&channel::run_receive_loop, this));
}

channel::~channel()
{
    // this implicitly kills send and receive loop
    // (both will fail with an exception)
    close();
    boost::thread::id i = boost::this_thread::get_id();
    if (i == m_send_loop.get_id())
    {
        m_receive_loop.join();
    }
    else if (i == m_receive_loop.get_id())
    {
        m_send_loop.join();
    }
    else
    {
        // wait for both threads to exit
        m_send_loop.join();
        m_receive_loop.join();
    }
}

} } // namespace hamcast::ipc
