/******************************************************************************\
 *  _   ___     ____  __               _                                      *
 * | | | \ \___/ /  \/  | ___ __ _ ___| |_                                    *
 * | |_| |\     /| |\/| |/ __/ _` / __| __|                                   *
 * |  _  | \ - / | |  | | (_| (_| \__ \ |_                                    *
 * |_| |_|  \_/  |_|  |_|\___\__,_|___/\__|                                   *
 *                                                                            *
 * This file is part of the HAMcast project.                                  *
 *                                                                            *
 * HAMcast is free software: you can redistribute it and/or modify            *
 * it under the terms of the GNU Lesser General Public License as published   *
 * by the Free Software Foundation, either version 3 of the License, or       *
 * (at your option) any later version.                                        *
 *                                                                            *
 * HAMcast is distributed in the hope that it will be useful,                 *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of             *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                       *
 * See the GNU Lesser General Public License for more details.                *
 *                                                                            *
 * You should have received a copy of the GNU Lesser General Public License   *
 * along with HAMcast. If not, see <http://www.gnu.org/licenses/>.            *
 *                                                                            *
 * Contact: HAMcast support <hamcast-support@informatik.haw-hamburg.de>       *
\******************************************************************************/

#include <set>
#include <iostream>

#include "session.hpp"
#include "middleware.hpp"
#include "tech_module.hpp"

namespace boost {

void intrusive_ptr_add_ref(hamcast::middleware::session* ptr)
{
    hamcast::ref_counted::add_ref(ptr);
}

void intrusive_ptr_release(hamcast::middleware::session* ptr)
{
    hamcast::ref_counted::release(ptr);
}

} // namespace boost

using boost::bind;
using namespace hamcast::ipc;

namespace hamcast { namespace middleware {

namespace {

void cb_cr_sock_stream(session* self,
                       ipc::stream_id& result,
                       socket_id sid,
                       const uri& group)
{
    HC_LOG_TRACE("sid = " << sid << ", group = " << group.str());
    HC_REQUIRE(!group.empty());
    result = self->socket(sid).add_stream(group);
    HC_LOG_DEBUG("stream_id (result) = " << result);
}

void cb_get_sock_ifs(session* self,
                     std::vector<interface_id>& result,
                     socket_id sid)
{
    HC_LOG_TRACE("sid = " << sid);
    const tech_module_vec& iv = self->socket(sid).interfaces();
    HC_REQUIRE(!iv.empty());
    result.clear();
    result.reserve(iv.size());
    for (tech_module_vec::const_iterator i(iv.begin()); i != iv.end(); ++i)
    {
        const tech_module::ptr& tm = *i;
        HC_REQUIRE(tm.get() != 0);
        result.push_back((*i)->iface_id());
    }
}

void cb_designated_host(session*,
                        bool& result,
                        interface_id iid,
                        const uri& group)
{
    HC_LOG_TRACE("iid = " << iid << ", group = " << group.str());
    tech_module::ptr tm = tech_module::instance_by_iid(iid);
    HC_REQUIRE_VERBOSE(tm.get() != 0,
                            "No such interface (id = " << iid << ")");
    result = tm->designated_host(group);
}

void cb_add_sock_if(session* self,
                    hamcast::util::unit&,
                    socket_id sid,
                    interface_id iid)
{
    HC_LOG_FATAL("Not implemented yet");
}

void cb_rm_sock_if(session* self,
                   hamcast::util::unit&,
                   socket_id sid,
                   interface_id iid)
{
    HC_LOG_FATAL("Not implemented yet");
}

void cb_set_sock_ifs(session* self,
                     hamcast::util::unit&,
                     socket_id sid,
                     std::vector<interface_id> ifs)
{
    HC_REQUIRE(!ifs.empty());
    // get pointers
    std::vector<tech_module::ptr> tms;
    for (std::vector<interface_id>::iterator i = ifs.begin();
         i != ifs.end();
         ++i)
    {
        tech_module::ptr tm = tech_module::instance_by_iid(*i);
        HC_REQUIRE(tm.get() != 0);
        tms.push_back(tm);
    }
    // delegate call
    self->socket(sid).set_interfaces(tms);
}

void cb_neighbor_set(session*,
                     std::vector<uri>& result,
                     interface_id iid)
{
    HC_LOG_TRACE("iid = " << iid);
    tech_module::ptr tm = tech_module::instance_by_iid(iid);
    HC_REQUIRE_VERBOSE(tm.get() != 0,
                            "No such interface (id = " << iid << ")");
    tm->neighbor_set(result);
}

void cb_children_set(session*,
                     std::vector<uri>& result,
                     interface_id iid,
                     const uri& group)
{
    HC_REQUIRE(!group.empty());
    HC_LOG_TRACE("iid = " << iid << ", group = " << group.str());
    tech_module::ptr tm = tech_module::instance_by_iid(iid);
    HC_REQUIRE_VERBOSE(tm.get() != 0,
                            "No such interface (id = " << iid << ")");
    tm->children_set(result, group);
}

template<typename K, typename V>
void set_from_map(const std::map<K,V>& container, const K& key, V& storage)
{
    typename std::map<K,V>::const_iterator i = container.find(key);
    if (i != container.end()) storage = i->second;
}

void cb_get_ifs(session*, std::vector<hamcast::interface_property>& result)
{
    static std::string s_if_name = "if_name";
    static std::string s_if_addr = "if_addr";
    static std::string s_if_tech = "if_tech";
    HC_LOG_TRACE("");
    std::vector<tech_module::ptr> mods = tech_module::instances();
    for (size_t i = 0; i < mods.size(); ++i)
    {
        tech_module& tm = *(mods[i]);
        hamcast::interface_property p;
        p.id = tm.iface_id();
        // default values
        p.name = "";
        p.address = "";
        p.technology = "";
        const string_map& sd = tm.service_discovery_results();
        set_from_map(sd, s_if_name, p.name);
        set_from_map(sd, s_if_addr, p.address);
        set_from_map(sd, s_if_tech, p.technology);
        result.push_back(p);
    }
}

void cb_cr_sock(session* self, socket_id& result)
{
    HC_LOG_TRACE("");
    result = self->add_new_socket();
    HC_LOG_DEBUG("result = " << result);
}

void cb_rm_sock(session* self, hamcast::util::unit&, socket_id sid)
{
    HC_LOG_TRACE("sid = " << sid);
    self->remove_socket(sid);
}

void cb_parent_set(session*,
                   std::vector<uri>& result,
                   interface_id iid,
                   const uri& group)
{
    HC_REQUIRE(!group.empty());
    HC_LOG_TRACE("iid = " << iid << ", group = " << group.str());
    tech_module::ptr tm = tech_module::instance_by_iid(iid);
    HC_REQUIRE_VERBOSE(tm.get() != 0,
                            "No such interface (id = " << iid << ")");
    tm->parent_set(result, group);
}

void cb_group_set(session*,
                  std::vector<std::pair<uri, boost::uint32_t> >& result,
                  interface_id iid)
{
    HC_LOG_TRACE("iid = " << iid);
    tech_module::ptr tm = tech_module::instance_by_iid(iid);
    HC_REQUIRE_VERBOSE(tm.get() != 0,
                            "No such interface (id = " << iid << ")");
    tm->group_set(result);
}

void cb_set_ttl(session* self,
                hamcast::util::unit&,
                socket_id sid,
                boost::uint8_t ttl_value)
{
    HC_LOG_TRACE("sid = " << sid << ", ttl = " << ttl_value);
    self->socket(sid).set_ttl(ttl_value);
}

void cb_leave(session* self,
              hamcast::util::unit&,
              socket_id sid,
              const uri& group)
{
    HC_REQUIRE(!group.empty());
    HC_LOG_TRACE("sid = " << sid << ", u = " << group.str());
    self->socket(sid).leave(group);
}

void cb_join(session* self,
             hamcast::util::unit&,
             socket_id sid,
             const uri& group)
{
    HC_REQUIRE(!group.empty());
    HC_LOG_TRACE("sid = " << sid << ", u = " << group.str());
    self->socket(sid).join(group);
}

void cb_enable_events(session* self)
{
    HC_LOG_TRACE("");
    self->enable_events();
}

void cb_disable_events(session* self)
{
    HC_LOG_TRACE("");
    self->disable_events();
}

} // namespace <anonymous>

void session::enable_events()
{
    if (m_subscribed_events == false)
    {
        session::ptr self(this);
        subscribe_events(self);
        m_subscribed_events = true;
    }
}

void session::disable_events()
{
    if (m_subscribed_events == true)
    {
        session::ptr self(this);
        unsubscribe_events(self);
        m_subscribed_events = false;
    }
}

session::session(const util::source::ptr& in, const util::sink::ptr& out)
    : ipc::channel(in, out), m_subscribed_events(false)
{
    m_cr_sock_stream  = bind(cb_cr_sock_stream,  this, _1, _2, _3);
    m_get_sock_ifs    = bind(cb_get_sock_ifs,    this, _1, _2);
    m_designated_host = bind(cb_designated_host, this, _1, _2, _3);
    m_add_sock_if     = bind(cb_add_sock_if,     this, _1, _2, _3);
    m_set_sock_ifs    = bind(cb_set_sock_ifs,    this, _1, _2, _3);
    m_rm_sock_if      = bind(cb_rm_sock_if,      this, _1, _2, _3);
    m_neighbor_set    = bind(cb_neighbor_set,    this, _1, _2);
    m_children_set    = bind(cb_children_set,    this, _1, _2, _3);
    m_get_ifs         = bind(cb_get_ifs,         this, _1);
    m_cr_sock         = bind(cb_cr_sock,         this, _1);
    m_rm_sock         = bind(cb_rm_sock,         this, _1, _2);
    m_parent_set      = bind(cb_parent_set,      this, _1, _2, _3);
    m_group_set       = bind(cb_group_set,       this, _1, _2);
    m_set_ttl         = bind(cb_set_ttl,         this, _1, _2, _3);
    m_leave           = bind(cb_leave,           this, _1, _2, _3);
    m_join            = bind(cb_join,            this, _1, _2, _3);
    m_enable_events   = bind(cb_enable_events,   this);
    m_disable_events  = bind(cb_disable_events,  this);
}

void session::send(const ipc::message::ptr& what)
{
    if (what) m_queue.push(new session_msg(what, outgoing_ipc_msg_t));
}

void session::create(const util::source::ptr& in, const util::sink::ptr& out)
{
    (new session(in, out))->run();
}

void session::on_exit(const std::string& err_str)
{
    HC_LOG_TRACE("err_str = " << err_str);
    disable_events();
    for (socket_map::iterator i = m_sockets.begin(); i != m_sockets.end(); ++i)
    {
        i->second->leave_all();
    }
    m_sockets.clear();
}

socket_id session::add_new_socket()
{
    socket_id new_sid = m_socket_id_generator.next();
    m_sockets.insert(socket_map::value_type(new_sid, new socket_proxy(new_sid,
                                                                      this)));
    return new_sid;
}

socket_proxy& session::socket(socket_id sid)
{
    socket_map::iterator i = m_sockets.find(sid);
    HC_REQUIRE_VERBOSE(i != m_sockets.end(),
                            "No such socket (socket id = " << sid << ")");
    return *(i->second);
}

void session::remove_socket(socket_id sid)
{
    socket_map::iterator i = m_sockets.find(sid);
    if (i != m_sockets.end())
    {
        i->second->leave_all();
        m_sockets.erase(i);
    }
}

session::~session()
{
    session_msg* sm;
    while ((sm = m_queue.try_pop()) != 0)
    {
        delete sm;
    }
}

typedef std::vector<tech_module::ptr> tech_module_vec;

template<typename T>
inline void flush_if_needed(util::serializer& s, T& q)
{
//	static const boost::posix_time::microseconds wtime(50);
//	if (!q.wait_for_data(wtime)) s.flush();
    if (!q.front()) s.flush();
}

void session::send_loop()
{
    HC_LOG_TRACE("");
    session::ptr self_ptr(this);
    util::serializer s(m_sink);
    for (;;)
    {
        session_msg* sm = m_queue.pop();
        switch (sm->type)
        {

         case receive_loop_killed_t:
            HC_LOG_DEBUG("receive loop killed ... quit");
            delete sm;
            return;

         case outgoing_ipc_msg_t:
            s << *(sm->msg);
            flush_if_needed(s, m_queue);
            //s.flush();
            break;

         case outgoing_mcast_pckt_t:
            sm->mcast_pckt.first->delivered(sm->mcast_pckt.second);
            ipc::message::serialize_as_message(s,
                                               sm->mcast_pckt.first->id(),
                                               sm->mcast_pckt.second);
            flush_if_needed(s, m_queue);
            break;

         case force_acks_t:
            {
                //HC_LOG_SCOPE("force_acks_t", "");
                bool ack_forced = false;
                for (socket_map::iterator i = m_sockets.begin();
                     i != m_sockets.end();
                     ++i)
                {
                    if (i->second->force_acks(s) && !ack_forced)
                    {
                        ack_forced = true;
                    }
                }
                //flush_if_needed(s, m_queue);
                if (ack_forced) s.flush();
                else flush_if_needed(s, m_queue);
            }
            break;

         case incoming_sync_request_t:
            switch (sync_request_view(*(sm->msg)).fun_id())
            {

             case fid_create_socket:
                ipc::create_socket.reply(sm->msg, s, m_cr_sock);
                break;

             case fid_delete_socket:
                ipc::delete_socket.reply(sm->msg, s, m_rm_sock);
                break;

             case fid_create_send_stream:
                ipc::create_send_stream.reply(sm->msg, s, m_cr_sock_stream);
                break;

             case fid_join:
                ipc::join.reply(sm->msg, s, m_join);
                break;

             case fid_leave:
                ipc::leave.reply(sm->msg, s, m_leave);
                break;

             case fid_set_ttl:
                ipc::set_ttl.reply(sm->msg, s, m_set_ttl);
                break;

             case fid_get_sock_interfaces:
                ipc::get_sock_interfaces.reply(sm->msg, s, m_get_sock_ifs);
                break;

             case fid_add_sock_interface:
                ipc::add_sock_interface.reply(sm->msg, s, m_add_sock_if);
                break;

             case fid_del_sock_interface:
                ipc::del_sock_interface.reply(sm->msg, s, m_rm_sock_if);
                break;

             case fid_set_sock_interfaces:
                ipc::set_sock_interfaces.reply(sm->msg, s, m_set_sock_ifs);
                break;

             case fid_get_interfaces:
                ipc::get_interfaces.reply(sm->msg, s, m_get_ifs);
                break;

             case fid_group_set:
                ipc::group_set.reply(sm->msg, s, m_group_set);
                break;

             case fid_neighbor_set:
                ipc::neighbor_set.reply(sm->msg, s, m_neighbor_set);
                break;

             case fid_children_set:
                ipc::children_set.reply(sm->msg, s, m_children_set);
                break;

             case fid_parent_set:
                ipc::parent_set.reply(sm->msg, s, m_parent_set);
                break;

             case fid_designated_host:
                ipc::designated_host.reply(sm->msg, s, m_designated_host);
                break;

             case fid_enable_events:
                ipc::enable_events.reply(sm->msg, s, m_enable_events);
                break;

             case fid_disable_events:
                ipc::disable_events.reply(sm->msg, s, m_disable_events);
                break;

             default: throw requirement_failed("Invalid function ID");

            }
            //flush_if_needed(s, m_queue);
            s.flush();
            break;

         case incoming_async_send_t:
            {
                ipc::async_send_view asv(*(sm->msg));
                HC_LOG_DEBUG("async_send"
                             << "; socket = " << asv.sck_id()
                             << ", stream = " << asv.stm_id()
                             << ", seq nr = " << asv.seq_nr());
                socket_map::iterator i = m_sockets.find(asv.sck_id());
                if (i != m_sockets.end())
                {
                    i->second->handle_async_send(asv, s, self_ptr);
                }
                else
                {
                    HC_LOG_DEBUG("Unknown socket: " << asv.sck_id());
                }
            }
            break;

         case async_send_job_done_t:
            {
                ipc::async_send_view asv(*(sm->msg));
                socket_map::iterator i = m_sockets.find(asv.sck_id());
                if (i != m_sockets.end())
                {
                    i->second->async_job_done(asv, s);
                }
                else
                {
                    HC_LOG_DEBUG("Unknown socket: " << asv.sck_id());
                }
                flush_if_needed(s, m_queue);
            }
            break;

         case membership_event_t:
            ipc::message::serialize_as_message(s, sm->mbs_event);
            flush_if_needed(s, m_queue);
            break;

        }
        delete sm;
    }
}

void session::receive_loop()
{
    util::deserializer d(m_source);
    try
    {
        for (;;)
        {
            while (!m_source->wait_for_data(0, force_ack_us_interval))
            {
                m_queue.push(new session_msg(force_acks()));
            }
            message::ptr msg;
            d >> msg;
            if (msg->type() == ipc::sync_request)
            {
                m_queue.push(new session_msg(msg, incoming_sync_request_t));
            }
            else if (msg->type() == ipc::async_send)
            {
                m_queue.push(new session_msg(msg, incoming_async_send_t));
            }
            else
            {
                std::ostringstream err;
                err << "Client sendet illegal message type: " << msg->type();
                std::string err_str = err.str();
                HC_LOG_FATAL(err_str);
                throw std::runtime_error(err_str);
            }
        }
    }
    catch (...)
    {
        m_queue.push(new session_msg(receive_loop_killed()));
        // rethrow exception (catched by channel::run_receive_loop)
        throw;
    }
}

} } // namespace hamcast::middleware
