/******************************************************************************\
 *  _   ___     ____  __               _                                      *
 * | | | \ \___/ /  \/  | ___ __ _ ___| |_                                    *
 * | |_| |\     /| |\/| |/ __/ _` / __| __|                                   *
 * |  _  | \ - / | |  | | (_| (_| \__ \ |_                                    *
 * |_| |_|  \_/  |_|  |_|\___\__,_|___/\__|                                   *
 *                                                                            *
 * This file is part of the HAMcast project.                                  *
 *                                                                            *
 * HAMcast is free software: you can redistribute it and/or modify            *
 * it under the terms of the GNU Lesser General Public License as published   *
 * by the Free Software Foundation, either version 3 of the License, or       *
 * (at your option) any later version.                                        *
 *                                                                            *
 * HAMcast is distributed in the hope that it will be useful,                 *
 * but WITHOUT ANY WARRANTY; without even the implied warranty of             *
 * MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.                       *
 * See the GNU Lesser General Public License for more details.                *
 *                                                                            *
 * You should have received a copy of the GNU Lesser General Public License   *
 * along with HAMcast. If not, see <http://www.gnu.org/licenses/>.            *
 *                                                                            *
 * Contact: HAMcast support <hamcast-support@informatik.haw-hamburg.de>       *
\******************************************************************************/

#ifndef SESSION_HPP
#define SESSION_HPP

#include <queue>
#include <vector>
#include <utility>
#include <boost/thread.hpp>
#include "hamcast/hamcast.hpp"

#include "middleware.hpp"
#include "connection.hpp"
#include "session_fwd.hpp"
#include "socket_proxy.hpp"

#include "hamcast/socket_id.hpp"
#include "hamcast/interface_id.hpp"
#include "hamcast/interface_property.hpp"

#include "hamcast/ipc/api.hpp"
#include "hamcast/ipc/channel.hpp"

#include "hamcast/util/id_generator.hpp"
#include "hamcast/util/const_buffer.hpp"
#include "hamcast/util/single_reader_queue.hpp"

namespace hamcast { namespace middleware {

enum session_msg_t
{
    outgoing_ipc_msg_t,
    outgoing_mcast_pckt_t,
    incoming_sync_request_t,
    incoming_async_send_t,
    async_send_job_done_t,
    force_acks_t,
    receive_loop_killed_t,
    membership_event_t
};

struct force_acks { };
struct receive_loop_killed { };

struct session_msg
{

    session_msg* next;
    session_msg_t type;
    ipc::message::ptr msg;
    std::pair<socket_proxy*, multicast_packet> mcast_pckt;
    membership_event mbs_event;

    inline session_msg(force_acks) : next(0), type(force_acks_t) { }

    inline session_msg(receive_loop_killed)
        : next(0), type(receive_loop_killed_t)
    {
    }

    inline session_msg(const ipc::message::ptr& mptr, session_msg_t msg_type)
        : next(0), type(msg_type), msg(mptr)
    {
    }

    inline session_msg(const std::pair<socket_proxy*, multicast_packet>& mp)
        : next(0), type(outgoing_mcast_pckt_t), mcast_pckt(mp)
    {
    }

    inline session_msg(const membership_event& me)
        : next(0), type(membership_event_t), mbs_event(me)
    {
    }

};

typedef std::map<socket_id, socket_proxy::ptr> socket_map;

/**
 * @brief Describes a client-specific session.
 *
 * A session consist of a socket_id_generator, an IPC connection and a map
 * of socket_proxys that the user created with create_socket().
 */
class session : public ipc::channel
{

    bool m_subscribed_events;

    socket_map m_sockets;
    util::id_generator<socket_id> m_socket_id_generator;
    util::single_reader_queue<session_msg> m_queue;

    // callbacks (ipc::sync_function implementation)
    ipc::create_send_stream_t::callback m_cr_sock_stream;
    ipc::get_sock_interfaces_t::callback m_get_sock_ifs;
    ipc::designated_host_t::callback m_designated_host;
    ipc::add_sock_interface_t::callback m_add_sock_if;
    ipc::del_sock_interface_t::callback m_rm_sock_if;
    ipc::set_sock_interfaces_t::callback m_set_sock_ifs;
    ipc::neighbor_set_t::callback m_neighbor_set;
    ipc::children_set_t::callback m_children_set;
    ipc::get_interfaces_t::callback m_get_ifs;
    ipc::create_socket_t::callback m_cr_sock;
    ipc::delete_socket_t::callback m_rm_sock;
    ipc::parent_set_t::callback m_parent_set;
    ipc::group_set_t::callback m_group_set;
    ipc::set_ttl_t::callback m_set_ttl;
    ipc::leave_t::callback m_leave;
    ipc::join_t::callback m_join;
    ipc::enable_events_t::callback m_enable_events;
    ipc::disable_events_t::callback m_disable_events;

    session(const util::source::ptr& in, const util::sink::ptr& out);

 protected:

    virtual void send_loop();

    virtual void receive_loop();

    virtual void on_exit(const std::string &err_str);

 public:

    /**
     * @brief A smart pointer to an instance of session.
     */
    typedef boost::intrusive_ptr<session> ptr;

    ~session();

    virtual void send(const ipc::message::ptr& what);

    inline void send(const std::pair<socket_proxy*, multicast_packet>& smp)
    {
        m_queue.push(new session_msg(smp));
    }

    inline void handle_event(const membership_event& mevent)
    {
        m_queue.push(new session_msg(mevent));
    }

    inline void async_job_done(const ipc::message::ptr& what)
    {
        m_queue.push(new session_msg(what,
                                     async_send_job_done_t));
    }

    /**
     * @brief Create a new session that uses @p io.
     * @note The created session runs asynchronously.
     * @pre @p io is connected to a verified (compatible) client.
     */
    static void create(const util::source::ptr& in,
                       const util::sink::ptr& out);

    /**
     * @brief Add a new socket to the session.
     */
    socket_id add_new_socket();

    /**
     * @brief Get the socket with id @p sid from the session.
     */
    socket_proxy& socket(socket_id sid);

    inline bool has_socket(socket_id sid)
    {
        return m_sockets.count(sid) != 0;
    }

    /**
     * @brief Remove the socket with id @p sid from this session.
     * @pre @p sid is a valid socket id for this session.
     */
    void remove_socket(socket_id sid);

    void enable_events();

    void disable_events();

};

} } // namespace hamcast::middleware

#endif // SESSION_HPP
